from typing import Any

from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from langchain_core.messages import HumanMessage, SystemMessage
from pydantic import BaseModel, Field
from wbcore.contrib.ai.exceptions import APIStatusErrors
from wbcore.contrib.ai.llm.utils import run_llm


class CompanyModel(BaseModel):
    name: str = Field(..., description="The name of the publicly listed company.")
    isin: str | None = Field(..., description="The ISIN of the publicly listed company. May be null if not available.")
    ticker: str | None = Field(
        ..., description="The ticker of the publicly listed company. May be null if not available."
    )
    refinitiv_identifier_code: str | None = Field(
        ..., description="The RIC of the publicly listed company. May be null if not available."
    )
    sentiment: int = Field(
        ...,
        ge=1,
        le=4,
        description="The sentiment of the news article towards the company. 1 is negative, 2 is neutral, 3 is positive, 4 is very positive.",
    )
    analysis: str = Field(
        ...,
        description="The analysis of the news article towards the company. This should include an explanation of the sentiment.",
    )


class CompaniesModel(BaseModel):
    companies: list[CompanyModel] = Field(
        default=[], description="The list of companies mentioned in the news article."
    )


@shared_task(
    autoretry_for=tuple(APIStatusErrors),
    retry_backoff=10,
    max_retries=5,  # retry i5 times maximum
    default_retry_delay=30,  # retry in 30s
    retry_jitter=False,
)
def run_company_extraction_llm(title: str, description: str, *args) -> list[dict[str, Any]]:
    from wbfdm.import_export.handlers.instrument import InstrumentLookup
    from wbfdm.models import Instrument

    res = run_llm(
        prompt=[
            SystemMessage(
                content="You will be parsed a news article, please provide the name of the publicly listed companies mentioned in the article, along with their ISIN, ticker, RIC, sentiment, and analysis."
            ),
            HumanMessage(content=f"Title: {title}, Description: {description}"),
        ],
        output_model=CompaniesModel,
    )
    relationships = []
    instrument_ct = ContentType.objects.get_for_model(Instrument)
    for company in res.get("companies", []):
        instrument = InstrumentLookup(Instrument).lookup(
            only_security=True,
            name=company.name,
            isin=company.isin,
            ticker=company.ticker,
            refinitiv_identifier_code=company.refinitiv_identifier_code,
        )
        if instrument is not None:
            relationships.append(
                {
                    "content_type_id": instrument_ct.id,
                    "object_id": instrument.get_root().id,
                    "sentiment": company.sentiment,
                    "analysis": company.analysis,
                }
            )

    return relationships



from datetime import date

from celery import shared_task
from django.db import transaction
from django.db.models import Q
from pandas.tseries.offsets import BDay
from tqdm import tqdm
from wbfdm.models import Instrument
from wbfdm.sync.runner import (  # noqa: F401
    initialize_exchanges,
    initialize_instruments,
    synchronize_exchanges,
    synchronize_instruments,
)


@shared_task(queue="portfolio")
def update_of_investable_universe_data(start: date | None = None, end: date | None = None, clear: bool = False):
    """
    Update the investable universe data on a daily basis.

    Parameters:
    - start (date | None): The start date for updating data. If None, defaults to three business days before 'end'.
    - end (date | None): The end date for updating data. If None, defaults to the current date.

    Notes:
    - The function resets the investable universe by marking all instruments as not in the investable universe.
    - It then updates all instruments marked as part of the investable universe.
    - If 'end' is not provided, it defaults to the current date.
    - If 'start' is not provided, it defaults to three business days before 'end'.

    Returns:
    None
    """
    if not end:
        end = (
            date.today() - BDay(1)
        ).date()  # we don't import today price in case the dataloader returns duplicates (e.g. DSWS)
    if not start:
        start = (end - BDay(3)).date()  # override three last day by default

    Instrument.investable_universe.update(
        is_investable_universe=True
    )  # ensure all the investable universe is marked as such

    instruments = Instrument.active_objects.filter(is_investable_universe=True, delisted_date__isnull=True).exclude(
        Q(is_managed=True)
        | Q(dl_parameters__market_data__path="wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader")
    )  # we exclude product and index managed to avoid circular import
    for instrument in tqdm(instruments, total=instruments.count()):
        instrument.import_prices(start=start, end=end, clear=clear)


@shared_task(queue="portfolio")
def synchronize_instruments_as_task():
    synchronize_instruments()


@shared_task(queue="portfolio")
def synchronize_exchanges_as_task():
    synchronize_exchanges()


@shared_task(queue="portfolio")
def full_synchronization_as_task():
    initialize_exchanges()
    initialize_instruments()
    with transaction.atomic():
        Instrument.objects.rebuild()  # rebuild MPTT tree



from contextlib import suppress
from functools import reduce
from itertools import batched
from datetime import date
from enum import Enum
from typing import Iterator
import pypika as pk
from pypika import Column, MSSQLQuery, functions as fn
from pypika.enums import SqlTypes, Order
from pypika.terms import ValueWrapper, LiteralValue

from django.db import ProgrammingError, connections
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall
from wbfdm.dataloaders.protocols import MarketDataProtocol
from wbfdm.contrib.qa.dataloaders.utils import create_table
from wbfdm.dataloaders.types import MarketDataDict
from wbfdm.enums import Frequency, MarketData


class DS2MarketData(Enum):
    OPEN = ("Open_", None)
    CLOSE = ("Close_", None)
    HIGH = ("High", None)
    LOW = ("Low", None)
    BID = ("Bid", None)
    ASK = ("Ask", None)
    VWAP = ("VWAP", None)
    VOLUME = ("Volume", None)
    MARKET_CAPITALIZATION = ("ConsolMktVal", 1_000_000)
    SHARES_OUTSTANDING = ("ConsolNumShrs", 1_000)


class DatastreamMarketDataDataloader(MarketDataProtocol, Dataloader):
    def market_data(
        self,
        values: list[MarketData] | None = [MarketData.CLOSE],
        from_date: date | None = None,
        to_date: date | None = None,
        exact_date: date | None = None,
        frequency: Frequency = Frequency.DAILY,
        target_currency: str | None = None,
        **kwargs,
    ) -> Iterator[MarketDataDict]:
        """Get market data for instruments.

        Args:
            queryset (QuerySet["Instrument"]): The queryset of instruments.
            values (list[MarketData]): List of values to include in the results.
            from_date (date | None): The starting date for filtering prices. Defaults to None.
            to_date (date | None): The ending date for filtering prices. Defaults to None.
            frequency (Frequency): The frequency of the requested data

        Returns:
            Iterator[MarketDataDict]: An iterator of dictionaries conforming to the DailyValuationDict.
        """

        lookup = {
            f"{k[0]},{k[1]}": v for k, v in self.entities.values_list("dl_parameters__market_data__parameters", "id")
        }
        value_mapping = [(DS2MarketData[x.name].value, x.value) for x in values or []]

        # Define tables
        pricing = pk.Table("vw_DS2Pricing")
        market_val = pk.Table("DS2MktVal")
        fx_code = pk.Table("DS2FxCode")
        fx = pk.Table("DS2FxRate")

        mapping, create_mapping_table = create_table(
            "#ds2infoexchcode", Column("InfoCode", SqlTypes.INTEGER), Column("ExchIntCode", SqlTypes.INTEGER)
        )

        # Base query to get data we always need unconditionally
        query = (
            pk.MSSQLQuery.from_(pricing)
            .select(
                fn.Concat(pricing.InfoCode, ",", pricing.ExchIntCode).as_("external_identifier"),
                fn.Concat(
                    pricing.InfoCode, ",", pricing.ExchIntCode, "_", fn.Cast(pricing.MarketDate, SqlTypes.DATE)
                ).as_("id"),
                fn.Cast(pricing.MarketDate, SqlTypes.DATE).as_("valuation_date"),
                ValueWrapper("qa-ds2").as_("source"),
            )
            .left_join(market_val)
            .on((pricing.InfoCode == market_val.InfoCode) & (pricing.MarketDate == market_val.ValDate))
            # We join on _codes, which removes all instruments not in _codes - implicit where
            .join(mapping)
            .on((pricing.InfoCode == mapping.InfoCode) & (pricing.ExchIntCode == mapping.ExchIntCode))
            .where(pricing.AdjType == 2)
            .orderby(pricing.MarketDate, order=Order.desc)
        )

        # If we need to convert to a target currency, we need the fx rate table and multiply all values with the fx rate
        if target_currency:
            query = (
                query.select(
                    ValueWrapper(target_currency).as_("currency"),
                    *[
                        (LiteralValue(value[0][0]) * fn.Coalesce(fx.midrate, 1)).as_(value[1])
                        for value in value_mapping
                    ],
                )
                .left_join(fx_code)
                .on(
                    (fx_code.FromCurrCode == target_currency)
                    & (fx_code.ToCurrCode == pricing.Currency)
                    & (fx_code.RateTypeCode == "SPOT")
                )
                .left_join(fx)
                .on((fx_code.ExRateIntCode == fx.ExRateIntCode) & (fx.ExRateDate == pricing.MarketDate))
            )
        else:
            query = query.select(
                pricing.Currency.as_("currency"),
                *[LiteralValue(value[0][0]).as_(value[1]) for value in value_mapping],
            )

        # Add conditional where clauses
        if from_date:
            query = query.where(pricing.MarketDate >= from_date)

        if to_date:
            query = query.where(pricing.MarketDate <= to_date)

        if exact_date:
            query = query.where(pricing.MarketDate == exact_date)

        with connections["qa"].cursor() as cursor:
            with suppress(ProgrammingError):
                cursor.execute(create_mapping_table.get_sql())
                for batch in batched(
                    self.entities.values_list("dl_parameters__market_data__parameters", flat=True), 1000
                ):
                    cursor.execute(reduce(lambda x, y: x.insert(y), batch, MSSQLQuery.into(mapping)).get_sql())

            cursor.execute(query.get_sql())

            for row in dictfetchall(cursor, MarketDataDict):
                row["instrument_id"] = lookup[row["external_identifier"]]
                yield row

            cursor.execute(MSSQLQuery.drop_table(mapping).get_sql())
